@@ -889,7 +889,7 @@ su_ask_password (GksuContext *context, gchar *prompt,
           msg = g_strdup_printf (_("<b><big>Enter the password of %s to run "
                                    "the application '%s'"
                                    "</big></b>"),
-				   command, context->user);
+				   context->user, command);
       }
 
       gksuui_dialog_set_message (GKSUUI_DIALOG(dialog), msg);
@@ -1348,6 +1348,7 @@ gksu_context_new ()
   context->message = NULL;
   context->alert = NULL;
   context->grab = TRUE;
+  context->always_ask_password = FALSE;
 
   context->debug = FALSE;
 
@@ -1620,6 +1621,38 @@ gksu_context_get_grab (GksuContext *context)
   return context->grab;
 }
 
+/**
+ * gksu_context_set_always_ask_password:
+ * @context: the #GksuContext you want to modify
+ * @value: TRUE or FALSE
+ *
+ * Set up if gksu should always ask for a password. Notice that this
+ * will only work when passwords are cached, as done by gnome-keyring
+ * for gksu's su mode and by sudo for gksu's sudo mode, but will have no
+ * effect if su or sudo are set up to not require the password at all.
+ */
+void
+gksu_context_set_always_ask_password (GksuContext *context, gboolean value)
+{
+  context->always_ask_password = value;
+}
+
+/**
+ * gksu_context_get_always_ask_password:
+ * @context: the #GksuContext you want to ask whether a grab will be done.
+ *
+ * Returns TRUE if gksu has been asked to always ask for a password 
+ * (even if sudo or gnome-keyring have cached it)
+ *
+ * Returns: TRUE if yes, FALSE otherwise.
+ */
+gboolean
+gksu_context_get_always_ask_password (GksuContext *context)
+{
+   return context->always_ask_password;
+}
+
+
 /**
  * gksu_context_set_launcher_context:
  * @context: the #GksuContext you want to set the sn context in
@@ -1995,8 +2028,12 @@ gksu_su_full (GksuContext *context,
 	      if (prompt_grab)
 		gksu_prompt_grab (context);
 
-	      /* try to get the password from the GNOME Keyring first */
-	      password = get_gnome_keyring_password (context);
+              /* try to get the password from the GNOME Keyring first, but
+	       * only if we have not been requested to always ask for the
+	       * password
+	       */
+	      if (!context->always_ask_password)
+	        password = get_gnome_keyring_password (context);
 	      if (password == NULL)
 		{
 		  password = ask_pass (context, buf, ask_pass_data, error);
@@ -2193,7 +2230,8 @@ gksu_su_full (GksuContext *context,
       cmdline = g_strdup("bin/su");
       /* wait for the child process to end or become something other
 	 than su */
-      while ((!waitpid (pid, &status, WNOHANG)) &&
+      pid_t pid_exited;
+      while ((!(pid_exited = waitpid (pid, &status, WNOHANG))) &&
 	     (g_str_has_suffix(cmdline, "bin/su")))
 	{
 	  if (cmdline)
@@ -2212,7 +2250,7 @@ gksu_su_full (GksuContext *context,
 	  bzero(buf, 256);
 	}
 
-      if (!WIFEXITED(status))
+      if (pid_exited != pid)
 	waitpid(pid, &status, 0);
 
       if (WEXITSTATUS(status))
@@ -2361,12 +2399,18 @@ gksu_sudo_full (GksuContext *context,
 	fprintf (stderr, "No ask_pass set, using default!\n");
       ask_pass = su_ask_password;
     }
-
   if (pass_not_needed == NULL)
     {
       pass_not_needed = no_pass;
     }
 
+  if (context->always_ask_password)
+    {
+       gint exit_status;
+       g_spawn_command_line_sync("/usr/bin/sudo -K", NULL, NULL, &exit_status, NULL);
+    }
+
+
   /*
      FIXME: need to set GError in a more detailed way
   */
@@ -2681,8 +2725,9 @@ gksu_sudo_full (GksuContext *context,
 
       cmdline = g_strdup("sudo");
       /* wait for the child process to end or become something other
-	 then sudo */
-      while ((!waitpid (pid, &status, WNOHANG)) &&
+	 than sudo */
+      pid_t pid_exited;
+      while ((!(pid_exited = waitpid (pid, &status, WNOHANG))) &&
 	     (g_str_has_suffix(cmdline, "sudo")))
 	{
 	  if (cmdline)
@@ -2700,7 +2745,8 @@ gksu_sudo_full (GksuContext *context,
 	  bzero(buffer, 256);
 	}
 
-      if (!WIFEXITED(status))
+      /* if the process is still active waitpid() on it */
+      if (pid_exited != pid)
 	waitpid(pid, &status, 0);
       sudo_reset_xauth (context, xauth, xauth_env);
 
